import 'dart:convert';

import 'dart:io';
import 'dart:math';
import 'dart:typed_data';

import 'package:expandable/expandable.dart';
import 'package:file_picker/file_picker.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter_datetime_picker/flutter_datetime_picker.dart';
import 'package:font_awesome_flutter/font_awesome_flutter.dart';
import 'package:path_provider/path_provider.dart';
// import 'package:pull_to_refresh/pull_to_refresh.dart';

import 'package:beevoter_common/model/controller.dart';
import 'package:pull_to_refresh/pull_to_refresh.dart';

class CommonConversion {
  static DateTime baseTime = DateTime.now();
  static String date2String(DateTime dt) {
    // 2019-09-05T00:00:00
    if (dt != null) {
      var ret = dt
          .subtract(new Duration(minutes: baseTime.timeZoneOffset.inMinutes))
          .toString();
      return ret;
    }
    return null;
  }

  static DateTime string2Date(dynamic dt) {
    // 2019-09-05T00:00:00
    if (dt != null) {
      var dtStr = dt.toString();
      if (dtStr.length > 0) {
        if (dtStr.length == 5) {
          var ret = DateTime.now();
          var hour = dtStr.substring(0, 2);
          var min = dtStr.substring(3, 5);
          return DateTime(ret.year, ret.month, ret.day, int.parse(hour),
              int.parse(min), 0)
              .add(new Duration(minutes: baseTime.timeZoneOffset.inMinutes));
        }
        var year = dtStr.substring(0, 4);
        var month = dtStr.substring(5, 7);
        var day = dtStr.substring(8, 10);
        if (dtStr.length > 11) {
          var hour = dtStr.substring(11, 13);
          var min = dtStr.substring(14, 16);
          var sec = dtStr.substring(17, 19);
          return DateTime(int.parse(year), int.parse(month), int.parse(day),
              int.parse(hour), int.parse(min), int.parse(sec))
              .add(new Duration(minutes: baseTime.timeZoneOffset.inMinutes));
        }
        return DateTime(
            int.parse(year), int.parse(month), int.parse(day), 0, 0, 0)
            .add(new Duration(minutes: baseTime.timeZoneOffset.inMinutes));
      }
    }
    return DateTime.now();
  }

  static String twoDigits(int d) {
    if (d < 10) return "0" + d.toString();
    return d.toString();
  }

  static String twoDigitsD(double d) {
    return d.toStringAsFixed(2);
  }

  static String formatDateAndTime(DateTime dt) {
    if (dt != null) {
      if (BeeVoterController.currentLanguage == 'pt-BR')
        return twoDigits(dt.day) +
            '/' +
            twoDigits(dt.month) +
            '/' +
            twoDigits(dt.year) +
            ' ' +
            twoDigits(dt.hour) +
            ':' +
            twoDigits(dt.minute);

      return twoDigits(dt.month) +
          '/' +
          twoDigits(dt.day) +
          '/' +
          twoDigits(dt.year) +
          ' ' +
          twoDigits(dt.hour) +
          ':' +
          twoDigits(dt.minute);
    }
    return "";
  }

  static String formatTime(DateTime dt) {
    if (dt != null) {
      return twoDigits(dt.hour) + ':' + twoDigits(dt.minute);
    }
    return "";
  }

  static String formatDate(DateTime dt) {
    if (dt != null) {
      if (BeeVoterController.currentLanguage == 'pt-BR')
        return twoDigits(dt.day) +
            '/' +
            twoDigits(dt.month) +
            '/' +
            twoDigits(dt.year);
      return twoDigits(dt.month) +
          '/' +
          twoDigits(dt.day) +
          '/' +
          twoDigits(dt.year);
    }
    return "";
  }

  static String toHexColor(Color color) {
    return '#'
        '${color.red.toRadixString(16).padLeft(2, '0')}'
        '${color.green.toRadixString(16).padLeft(2, '0')}'
        '${color.blue.toRadixString(16).padLeft(2, '0')}'
        '${color.alpha.toRadixString(16).padLeft(2, '0')}';
  }

  static Color randomColorGenerator() {
    var r = new Random();
    return Color.fromRGBO(r.nextInt(256), r.nextInt(256), r.nextInt(256), 0.75);
  }

  static Color fromHexColor(String hexString, Color defaultColor) {
    final buffer = StringBuffer();
    if (hexString == null || hexString.isEmpty) return defaultColor;
    if (hexString == "white") return DefaultTheme.cardColor;
    if (hexString == "black") return DefaultTheme.textColor;
    if (hexString.indexOf("rgba") == 0) {
      var colors = hexString.substring("rgba(".length);
      colors = colors.substring(0, colors.length - 1);
      var sColors = colors.split(',');
      var red = int.parse(sColors[0]);
      var green = int.parse(sColors[1]);
      var blue = int.parse(sColors[2]);
      var opacity = double.parse(sColors[3]);
      return Color.fromRGBO(red, green, blue, opacity);
    }

    buffer.write(hexString.replaceFirst('#', ''));
    try {
      var c = buffer.toString();
      if (hexString.length == 6 || hexString.length == 7)
        c = 'ff' + c;
      else
        c = c.substring(6, 8) + c.substring(0, 6);
      return Color(int.parse(c, radix: 16));
    } catch (message) {
      return defaultColor;
    }
  }

  static showAlertDialog(
      BuildContext context, String title, String message, VoidCallback onOk) {
    // set up the button
    Widget okButton = TextButton(
      child: Text("OK"),
      onPressed: () {
        Navigator.of(context).pop(); // dismiss dialog
        onOk();
      },
    );

    // set up the AlertDialog
    AlertDialog alert = AlertDialog(
      title: Text(title),
      content: Text(message),
      actions: [
        okButton,
      ],
    );

    // show the dialog
    showDialog(
      context: context,
      builder: (BuildContext context) {
        return alert;
      },
    );
  }

  static showYesNoDialog(BuildContext context, String title, String message,
      VoidCallback onYes, VoidCallback onNo) {
    // set up the buttons
    Widget cancelButton = TextButton(
      child: Text(
        BeeVoterController.translate("No"),
        style: TextStyle(
          color: Colors.black,
          fontSize: DefaultTheme.fontSmall,
        ),
      ),
      onPressed: () {
        Navigator.of(context).pop(); // dismiss dialog
        onNo();
      },
    );
    Widget continueButton = TextButton(
      child: Text(BeeVoterController.translate("Yes"),
          style: TextStyle(
            color: Colors.black,
            fontSize: DefaultTheme.fontSmall,
          )),
      onPressed: () {
        Navigator.of(context).pop(); // dismiss dialog
        onYes();
      },
    );

    // set up the AlertDialog
    AlertDialog alert = AlertDialog(
      titlePadding: EdgeInsets.all(10),
      contentPadding: EdgeInsets.all(10),
      title: Text(title),
      content: Text(message),
      actions: [
        cancelButton,
        continueButton,
      ],
    );

    // show the dialog
    showDialog(
      context: context,
      builder: (BuildContext context) {
        return alert;
      },
    );
  }

  static showDeleteConfirm(BuildContext context, VoidCallback onYes) {
    showYesNoDialog(
        context,
        BeeVoterController.translate("Do you want to delete this record ?"),
        BeeVoterController.translate("You will not be able to revert this !"),
        onYes,
            () => {});
  }

//   static changeColorDialog(BuildContext context, Color pickerColor,
//       void Function(Color color) changeColor) {
//     showDialog(
//       context: context,
//       child: AlertDialog(
//         title: const Text('Pick a color!'),
//         content: SingleChildScrollView(
// //          child: ColorPicker(
// //            pickerColor: pickerColor,
// //            onColorChanged: changeColor,
// //            enableLabel: true,
// //            pickerAreaHeightPercent: 0.8,
// //          ),
//           // Use Material color picker:
//           //
//           child: MaterialPicker(
//             pickerColor: pickerColor,
//             onColorChanged: (Color color) {
//               changeColor(color);
//               Navigator.of(context).pop();
//             },
//             enableLabel: true, // only on portrait mode
//           ),
//           //
//           // Use Block color picker:
//           //
//           // child: BlockPicker(
//           //   pickerColor: currentColor,
//           //   onColorChanged: changeColor,
//           // ),
//         ),
//         actions: <Widget>[
//           FlatButton(
//             child: const Text('Select'),
//             onPressed: () {
//               changeColor(pickerColor);
//               Navigator.of(context).pop();
//             },
//           ),
//         ],
//       ),
//     );
//   }

  static bool isInside(DateTime date, DateTime from, DateTime to) {
    var difFrom = date.difference(from);
    var difTo = date.difference(to);
    return difFrom.inDays >= 0 && (difTo.inSeconds == 0 || difTo.isNegative);
  }

  static List<DateTime> getListOfDays(DateTime from, DateTime to) {
    if (to == null && from == null) return [];
    if (to == null) return [from];
    if (from == null) return [to];

    var ret = []; // new List<DateTime>();
    var start = new DateTime(from.year, from.month, from.day, 0, 0, 0, 0);
    int cnt = 0;

    // print('between ${formatDate(from)} and ${formatDate(to)}');
    while (start.difference(to).inMilliseconds <= 0) {
      // print('${formatDate(start)}');
      ret.add(start);
      start = start.add(new Duration(days: 1));
      if (cnt++ > 100) break;
    }
    return ret;
  }

  static bool isEmail(String em) {
    String p =
        r'^(([^<>()[\]\\.,;:\s@\"]+(\.[^<>()[\]\\.,;:\s@\"]+)*)|(\".+\"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$';

    RegExp regExp = new RegExp(p);

    return regExp.hasMatch(em);
  }
}

class CommonBeeVoterBar extends AppBar {
  CommonBeeVoterBar(String title)
      : super(
    title: Row(
      mainAxisAlignment: MainAxisAlignment.spaceBetween,
      children: <Widget>[
        Text(
          title,
          style: TextStyle(
            color: DefaultTheme.barTitleFontColor,
            fontSize: DefaultTheme.fontSmall,
          ),
        ),
        Image(
          image: AssetImage('assets/bee_logo.png'),
          height: 20,
        ),
      ],
    ),
  );
}

class CommonBeeVoterScaffold extends Scaffold {
  static Color bColor = DefaultTheme.backgroundColor; // yellowAccent.shade100;

  CommonBeeVoterScaffold(String title, List<Widget> children,
      {Widget floatingActionButton,
        GlobalKey<ScaffoldState> scaffoldKey,
        Widget drawer})
      : super(
    appBar: CommonBeeVoterBar(title),
    key: scaffoldKey,
    backgroundColor: bColor,
    floatingActionButton: floatingActionButton,
    body: CommonWaitable(
      SingleChildScrollView(
          child: CommonWaitable(
            Padding(
              padding: EdgeInsets.all(10),
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.stretch,
                mainAxisAlignment: MainAxisAlignment.center,
                children: children,
              ),
            ),
          )),
    ),
    drawer: drawer,
  );
}

class CommonTextStyle extends TextStyle {
  CommonTextStyle()
      : super(
    color: Colors.orange[700],
    fontSize: DefaultTheme.fontMedium,
  );
}

const MaterialColor backgroundApp = MaterialColor(
  _backgroundAppValue,
  <int, Color>{
    50: Color(_backgroundAppValue),
    100: Color(_backgroundAppValue),
    200: Color(_backgroundAppValue),
    300: Color(_backgroundAppValue),
    400: Color(_backgroundAppValue),
    500: Color(_backgroundAppValue),
    600: Color(_backgroundAppValue),
    700: Color(_backgroundAppValue),
    800: Color(_backgroundAppValue),
    900: Color(_backgroundAppValue),
  },
);
const int _backgroundAppValue = 0xFFFFF9C4;

const MaterialColor foregroundApp = MaterialColor(
  _foregroundAppValue,
  <int, Color>{
    50: Color(_foregroundAppValue),
    100: Color(_foregroundAppValue),
    200: Color(_foregroundAppValue),
    300: Color(_foregroundAppValue),
    400: Color(_foregroundAppValue),
    500: Color(_foregroundAppValue),
    600: Color(_foregroundAppValue),
    700: Color(_foregroundAppValue),
    800: Color(_foregroundAppValue),
    900: Color(_foregroundAppValue),
  },
);
const int _foregroundAppValue = 0xFF3E2723;

class DefaultTheme {
  static double fontXSmall = 10;
  static double fontSmall = 12;
  static double fontMedium = 14;
  static double fontLarge = 16;
  static double fontXlarge = 20;

  static String currentFont = 'Normal';

  static setFont(String font) async {
    currentFont = font;
    switch (font) {
      case 'Small':
        setSmallFonts();
        break;
      case 'Large':
        setLargeFonts();
        break;
      default:
        setNormalFonts();
        break;
    }
    await BeeVoterController.refreshPref();
  }

  static setSmallFonts() {
    fontXSmall = 8;
    fontSmall = 10;
    fontMedium = 12;
    fontLarge = 14;
    fontXlarge = 18;
    currentFont = 'Small';
  }

  static setNormalFonts() {
    fontXSmall = 10;
    fontSmall = 12;
    fontMedium = 14;
    fontLarge = 16;
    fontXlarge = 20;
    currentFont = 'Normal';
  }

  static setLargeFonts() {
    fontXSmall = 14;
    fontSmall = 16;
    fontMedium = 18;
    fontLarge = 20;
    fontXlarge = 24;
    currentFont = 'Large';
  }

  static String currentTheme = 'Light';
  static setTheme(String theme) async {
    currentTheme = theme;
    if (theme == 'Dark')
      setDarkTheme();
    else
      setNormalTheme();
    await BeeVoterController.refreshPref();
  }

  static void Function() changeTheme;

  static setNormalTheme() {
    textColor = Colors.black;
    backgroundColor = Colors.white;
    menuColor = Colors.grey[300];
    destakListColor = Colors.yellow[100];
    cardColor = backgroundColor;
    removedCardColor = Colors.orange[100];
    removedCardTextColor = Colors.orange[300];

    folderBGColor = Colors.orangeAccent[100];
    folderFNColor = Colors.brown[900];

    barTitleFontColor = foregroundApp;
    barTitleBGColor = backgroundApp;
    publicTitleFontColor = Colors.brown[900];

    map3FNColor = Colors.brown[900];
    map3BGColor = Colors.orange[200];

    map5FNColor = Colors.yellow;
    map5BGColor = Colors.brown[800];

    map7FNColor = Colors.brown[900];
    map7BGColor = Colors.yellow[600];

    map8FNColor = Colors.brown[900];
    map8BGColor = Colors.orange;

    map9FNColor = Colors.brown[900];
    map9BGColor = Colors.greenAccent[400];

    buttomDeleteFNColor = map8FNColor;
    buttomDeleteBGColor = map8BGColor;

    buttomAddFNColor = map7FNColor;
    buttomAddBGColor = map7BGColor;

    buttomSaveFNColor = map5FNColor;
    buttomSaveBGColor = map5BGColor;

    if (changeTheme != null) changeTheme();
  }

  static setDarkTheme() {
    textColor = Colors.grey[300];
    backgroundColor = Colors.black;
    menuColor = Colors.grey[900];
    destakListColor = Colors.grey[850];
    removedCardColor = Colors.grey[850];
    removedCardTextColor = Colors.grey[600];

    folderBGColor = Colors.grey[800];
    folderFNColor = Colors.white;

    barTitleFontColor = foregroundApp;
    barTitleBGColor = backgroundApp;
    publicTitleFontColor = Colors.brown[900];

    map3FNColor = Colors.brown[900];
    map3BGColor = Colors.orange[200];

    map5BGColor = Colors.orange[800];
    map5FNColor = Colors.brown[800];

    map7FNColor = Colors.brown[900];
    map7BGColor = Colors.yellow[600];

    map8FNColor = Colors.brown[900];
    map8BGColor = Colors.orange;

    map9FNColor = Colors.brown[900];
    map9BGColor = Colors.greenAccent[400];

    buttomDeleteFNColor = map8FNColor;
    buttomDeleteBGColor = map8BGColor;

    buttomAddFNColor = map7FNColor;
    buttomAddBGColor = map7BGColor;

    buttomSaveFNColor = map5FNColor;
    buttomSaveBGColor = map5BGColor;

    if (changeTheme != null) changeTheme();
  }

  static Color textColor = Colors.black;
  static Color menuColor = Colors.grey[300];
  static Color backgroundColor = Colors.white;
  static Color destakListColor = Colors.yellow[100];
  static Color cardColor = Colors.white;
  static Color cardTextColor = Colors.black;
  static Color removedCardColor = Colors.orange[100];
  static Color removedCardTextColor = Colors.orange[300];

  static Color folderBGColor = Colors.orangeAccent[100];
  static Color folderFNColor = Colors.brown[900];

  static Color barTitleFontColor = foregroundApp;
  static Color barTitleBGColor =
      backgroundApp; //Colors.yellow[100]; // WARNING ... DONT USE [] COLORS
  static Color publicTitleFontColor = Colors.brown[900];

  static Color map3FNColor = Colors.brown[900];
  static Color map3BGColor = Colors.orange[200];

  static Color map5FNColor = Colors.yellow;
  static Color map5BGColor = Colors.brown[800];

  static Color map7FNColor = Colors.brown[900];
  static Color map7BGColor = Colors.yellow[600];

  static Color map8FNColor = Colors.brown[900];
  static Color map8BGColor = Colors.orange;

  static Color map9FNColor = Colors.brown[900];
  static Color map9BGColor = Colors.greenAccent[400];

  static Color buttomDeleteFNColor = map8FNColor;
  static Color buttomDeleteBGColor = map8BGColor;

  static Color buttomAddFNColor = map7FNColor;
  static Color buttomAddBGColor = map7BGColor;

  static Color buttomSaveFNColor = map5FNColor;
  static Color buttomSaveBGColor = map5BGColor;
}

class CommonList10TextStyle extends TextStyle {
  CommonList10TextStyle({Color color})
      : super(
    color: color == null ? DefaultTheme.textColor : color,
    fontSize: DefaultTheme.fontXSmall,
  );
}

class CommonList12TextStyle extends TextStyle {
  CommonList12TextStyle({Color color})
      : super(
    color: color == null ? DefaultTheme.textColor : color,
    fontSize: DefaultTheme.fontSmall,
  );
}

class CommonList14TextStyle extends TextStyle {
  CommonList14TextStyle({Color color})
      : super(
    color: color == null ? DefaultTheme.textColor : color,
    fontSize: DefaultTheme.fontMedium,
  );
}

class CommonIcon12 extends Icon {
  CommonIcon12(IconData icon, {Color color})
      : super(icon,
      size: DefaultTheme.fontSmall,
      color: color == null ? DefaultTheme.textColor : color);
}

class CommonIcon14 extends Icon {
  CommonIcon14(IconData icon, {Color color})
      : super(icon,
      size: DefaultTheme.fontMedium,
      color: color == null ? DefaultTheme.textColor : color);
}

class CommonIcon20 extends Icon {
  CommonIcon20(IconData icon, {Color color})
      : super(icon,
      size: DefaultTheme.fontXlarge,
      color: color == null ? DefaultTheme.textColor : color);
}

class CommonMessage extends Text {
  CommonMessage(String text)
      : super(text,
      style:
      TextStyle(color: Colors.red, fontSize: DefaultTheme.fontSmall));
}

class CommonButton extends StatelessWidget {
  final String title;
  final VoidCallback onPressed;

  CommonButton(this.title, this.onPressed);

//  @override
//  _CommonButtonState createState() => _CommonButtonState();
//}
//
//class _CommonButtonState extends State<CommonButton> {
  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return ButtonTheme(
      child: ElevatedButton(
        onPressed: onPressed,
        child: Text(
          title,
          style: TextStyle(
            color: DefaultTheme.map8FNColor,
            backgroundColor: DefaultTheme.map8BGColor,
            fontSize: DefaultTheme.fontMedium,
          ),
        ),
        // color: DefaultTheme.map8BGColor,
      ),
    );
  }
}

typedef BoolCallback = bool Function();

const EdgeInsetsGeometry initialCommonIconButtonPadding = EdgeInsets.all(2);

class CommonIconButton extends StatefulWidget {
  final IconData icon;
  final VoidCallback onPressed;
  final String tooltip;
  final Color fontColor;
  final Color backgroundColor;
  final bool isEnabled;
  final EdgeInsetsGeometry padding;

  CommonIconButton(this.icon, this.onPressed,
      {this.tooltip,
        this.fontColor,
        this.backgroundColor,
        this.isEnabled = true,
        this.padding = initialCommonIconButtonPadding});

  @override
  _CommonIconButtonState createState() => _CommonIconButtonState();
}

class _CommonIconButtonState extends State<CommonIconButton> {
  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return Container(
      alignment: Alignment.centerRight,
      padding: widget.padding,
      width: 40,
      height: 40,
      child: Container(
        decoration: new BoxDecoration(
          color: widget.isEnabled
              ? widget.backgroundColor == null
              ? DefaultTheme.map8BGColor
              : widget.backgroundColor
              : widget.backgroundColor == null
              ? DefaultTheme.map8BGColor.withOpacity(0.2)
              : widget.backgroundColor.withOpacity(0.2),
          borderRadius: BorderRadius.all(Radius.circular(100.0)),
        ),
        child: IconButton(
          color: widget.isEnabled
              ? widget.fontColor == null
              ? DefaultTheme.map8FNColor
              : widget.fontColor
              : widget.fontColor == null
              ? DefaultTheme.map8FNColor.withOpacity(0.2)
              : widget.fontColor.withOpacity(0.2),
          icon: CommonIcon14(
            widget.icon,
            color: widget.fontColor,
          ),
          onPressed: () {
            if (widget.isEnabled) widget.onPressed();
          },
          tooltip: widget.tooltip,
        ),
      ),
    );
  }
}

class CommonIconAndTextButton extends StatelessWidget {
  final String title;
  final IconData icon;
  final Color fontColor;
  final Color backgroundColor;
  final VoidCallback onPressed;

  CommonIconAndTextButton(this.title, this.icon, this.fontColor, this.backgroundColor, this.onPressed);

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return ButtonTheme(
      child: ElevatedButton(
        style: TextButton.styleFrom(
          backgroundColor: backgroundColor
        ),
        onPressed: onPressed,
        // color: backgroundColor,
        child: Wrap(
          children: <Widget>[
            Icon(icon, size: DefaultTheme.fontSmall, color: fontColor),
            Padding(
              padding: EdgeInsets.only(left: 5),
              child: Text(title,
                  style: TextStyle(
                    color: fontColor,
                    fontSize: DefaultTheme.fontMedium,
                  )),
            ),
          ],
        ),
      ),
    );
  }
}

class CommonFloatButton extends StatefulWidget {
  final String tooltip;
  final Icon icon;
  final VoidCallback onPressed;
  final BoolCallback isEnabled;
  final Color backgroundColor;

  callAction() {
    if (isEnabled == null || isEnabled()) onPressed();
  }

  CommonFloatButton(
      {this.tooltip,
        this.onPressed,
        this.icon,
        this.isEnabled,
        this.backgroundColor});

  _CommonFloatButtonState state;
  verifyOk(BuildContext context) {
    state.verifyOk();
  }

  @override
  _CommonFloatButtonState createState() => state = _CommonFloatButtonState();
}

class _CommonFloatButtonState extends State<CommonFloatButton> {
  bool lastEnabledPosition = false;

  verifyOk() {
    setState(() {
//      lastEnabledPosition = (widget.isEnabled == null || widget.isEnabled());
    });
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    lastEnabledPosition = (widget.isEnabled == null || widget.isEnabled());

    return FloatingActionButton(
      onPressed: widget.callAction,
      tooltip: widget.tooltip,
      child: widget.icon,
      materialTapTargetSize: MaterialTapTargetSize.shrinkWrap,
      foregroundColor: lastEnabledPosition
          ? DefaultTheme.textColor
          : DefaultTheme.textColor.withOpacity(0.1),
      backgroundColor: lastEnabledPosition
          ? Colors.orange
          : Colors.orange.withOpacity(0.2), // [200] ,
      elevation: 0,
    );
  }
}

class CommonAnchor extends Center {
  CommonAnchor(String title, GestureTapCallback onTap)
      : super(
    child: Padding(
      padding: EdgeInsets.only(bottom: 13),
      child: RichText(
        text: new TextSpan(
            text: title,
            style: CommonTextStyle(),
            recognizer: new TapGestureRecognizer()..onTap = onTap),
      ),
    ),
  );
}

class CommonSelectIdeableOption {
  int id;
  String title;

  CommonSelectIdeableOption(this.id, this.title);

  @override
  String toString() {
    return title;
  }
}

class CommonSelect<T> extends StatefulWidget {
  final String title;
  final List<T> items;
  final T value;
  final void Function(dynamic value) onChanged;

  CommonSelect(this.title, this.items, this.value, this.onChanged);

  @override
  _CommonSelectState<T> createState() => _CommonSelectState<T>(value);
}

class _CommonSelectState<T> extends State<CommonSelect> {
  T value;
  _CommonSelectState(T value) {
    this.value = value;
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return Container(
      // color: Colors.red,
      child: Row(children: <Widget>[
        Text(BeeVoterController.translate(widget.title) + ': ',
            style: CommonList10TextStyle()),
        DropdownButton(
          value: value,
          items: widget.items
              .map((i) => new DropdownMenuItem<T>(
              value: i, child: new Text(i.toString())))
              .toList(),
          onChanged: (value) {
            setState(() {
              this.value = value;
            });
            widget.onChanged(value);
          },
          style: CommonList12TextStyle(),
          iconEnabledColor: DefaultTheme.textColor.withOpacity(0.8),
        ),
      ]),
    );
  }
}

class CommonRadio<T> extends StatefulWidget {
  final String title;
  final List<T> items;
  final T value;
  final void Function(dynamic value) onChanged;

  CommonRadio(this.title, this.items, this.value, this.onChanged);

  @override
  _CommonRadioState<T> createState() => _CommonRadioState<T>(items, value);
}

class _CommonRadioState<T> extends State<CommonRadio> {
  _CommonRadioState(List<T> items, T selected) {
    this.selected = selected;
    if (this.selected == null)
      this.selected = items[0];
  }
  T selected;

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    var options = <Widget>[];
    options.add(Text(BeeVoterController.translate(widget.title) + ': ',
        style: CommonList10TextStyle()));
    for (var item in widget.items) {
      options.add(new RadioListTile<T>(
        activeColor: DefaultTheme.textColor,
        title: Text(item.toString(), style: CommonList12TextStyle()),
        value: item,
        groupValue: selected,
        onChanged: (T value) {
          setState(() {
            selected = value;
            widget.onChanged(value);
          });
        },
      ));
    }

    return Container(
//          width: 300.0,
      child: Column(children: options),
//        ),
    );
  }
}

class CommonDateField extends TextButton {
  CommonDateField(BuildContext context, String title, DateTime value,
      void Function(DateTime) onChange,
      {BuildContext Function() getContext})
      : super(
      onPressed: () {
        DatePicker.showDatePicker(
            getContext == null ? context : getContext(),
            showTitleActions: true,
            minTime: DateTime(1900, 1, 1),
            maxTime: DateTime(2100, 1, 1), onChanged: (date) {
          print('change $date');
        },
            onConfirm: onChange,
            currentTime: value,
            locale: BeeVoterController.localeType);
      },
      child: Text(
        title,
        style: CommonList12TextStyle(),
      ));
}

class CommonDateTimeField extends TextButton {
  CommonDateTimeField(BuildContext context, String title, DateTime value,
      void Function(DateTime) onChange,
      {BuildContext Function() getContext})
      : super(
      onPressed: () {
        DatePicker.showDateTimePicker(
            getContext == null ? context : getContext(),
            showTitleActions: true,
            minTime: DateTime(1900, 1, 1),
            maxTime: DateTime(2100, 1, 1), onChanged: (date) {
          print('change $date');
        },
            onConfirm: onChange,
            currentTime: value,
            locale: BeeVoterController.localeType);
      },
      child: Text(
        title,
        style: CommonList12TextStyle(),
      ));
}

class CommonTimeField extends TextButton {
  CommonTimeField(BuildContext context, String title, DateTime value,
      void Function(DateTime) onChange,
      {BuildContext Function() getContext})
      : super(
      onPressed: () {
        DatePicker.showTimePicker(
            getContext == null ? context : getContext(),
            showTitleActions: true, onChanged: (date) {
          print('change $date');
        },
            onConfirm: onChange,
            currentTime: value,
            locale: BeeVoterController.localeType);
      },
      child: Text(
        title,
        style: CommonList12TextStyle(),
      ));
}

class CommonCheckBoxAtLogin extends CheckboxListTile {
  CommonCheckBoxAtLogin(String title, bool value, ValueChanged<bool> onChanged)
      : super(
    title: Text(
      BeeVoterController.translate(title),
      style: CommonTextStyle(),
    ),
    value: value,
    onChanged: onChanged,
  );
}

class CommonSwitch extends StatefulWidget {
  final String title;
  final bool value;
  final ValueChanged<bool> onChanged;
  CommonSwitch(this.title, this.value, this.onChanged);

  @override
  _CommonSwitchState createState() => _CommonSwitchState(value);
}

class _CommonSwitchState extends State<CommonSwitch> {
  bool bValue;
  _CommonSwitchState(bool value) {
    this.bValue = value;
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return Wrap(
      children: <Widget>[
        Padding(
          padding: EdgeInsets.only(right: 5),
          child: Text(widget.title, style: CommonList12TextStyle()),
        ),
        Switch(
          value: bValue,
          onChanged: (value) {
            setState(() {
              this.bValue = value;
              widget.onChanged(value);
            });
          },
        ),
      ],
    );
  }
}

class CommonCheckBox extends StatefulWidget {
  final String title;
  final bool value;
  final ValueChanged<bool> onChanged;
  CommonCheckBox(this.title, this.value, this.onChanged);

  @override
  _CommonCheckBoxState createState() => _CommonCheckBoxState(value);
}

class _CommonCheckBoxState extends State<CommonCheckBox> {
  bool value;

  @override
  _CommonCheckBoxState(bool value) {
    this.value = value;
  }

  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return CheckboxListTile(
      title: Text(
        BeeVoterController.translate(widget.title),
        style: CommonList12TextStyle(),
      ),
      value: value,
      onChanged: (value) {
        setState(() {
          this.value = value;
          widget.onChanged(value);
        });
      },
      checkColor: DefaultTheme.barTitleFontColor,
      activeColor: DefaultTheme.barTitleBGColor,
    );
  }
}

//class CommonCheckBox extends CheckboxListTile {
//  CommonCheckBox(String title, bool value, ValueChanged<bool> onChanged)
//      : super(
//          title: Text(
//            BeeVoterController.translate(title),
//            style: CommonList12TextStyle(),
//          ),
//          value: value,
//          onChanged: onChanged,
//          checkColor: DefaultTheme.barTitleFontColor,
//        );
//}

class CommonFileInfo {
  String name;
  String type;
  Uint8List content;

  CommonFileInfo(this.name, this.type, this.content);

  //'taskAttachment', 'columnValue'
  static Future<CommonFileInfo> getFileContent(int id, String currentFileName,
      String currentType, String currentFileContent, String callType) async {
    if (id != null && id > 0) {
      String content = currentFileContent;
      if (content == null || content.isEmpty) {
        var url = 'GetStreamResourceByName?id=$id&type=$callType&guid=';
        var resp = await BeeVoterController.get(url);
//        content = resp; // utf8.decode(resp);
        var posB = currentFileName.lastIndexOf('/');
        var name = currentFileName.substring(posB + 1);
        var posE = name.lastIndexOf('.');
        var extension = name.substring(posE + 1).toLowerCase();
        var type = '';
        if (extension == 'jpg')
          type = 'image/jpeg';
        else if (extension == 'png')
          type = 'image/png';
        else if (extension == 'gif')
          type = 'image/gif';
        else if (extension == 'bmp')
          type = 'image/bmp';
        else if (extension == 'mp3')
          type = 'audio/mpeg';
        else if (extension == 'wav')
          type = 'audio/wav';
        else if (extension == 'txt')
          type = 'text/plain';
        else if (extension == 'pdf')
          type = 'application/pdf';
        else
          type = 'application/$extension';
        return new CommonFileInfo(name, type, resp);
      }
      if (content != null && content.isNotEmpty) {
        var jsContent = jsonDecode(content);
        var file = jsContent["file"];
        var name = file["name"];
//        var size = file["size"];
        var type = file["type"];
        var image = jsContent["image"].toString();
        var posB64 = image.indexOf('base64,') + 'base64,'.length;
        return new CommonFileInfo(
            name, type, base64.decode(image.substring(posB64)));
      }
    }
    return null;
  }
}

class CommonFileField extends StatefulWidget {
  final String title;
  final CommonFileInfo fileInfo;
  final void Function(String fileName, String fileContent) onSelect;

  CommonFileField(this.title, this.fileInfo, this.onSelect);

  @override
  _CommonFileFieldState createState() => _CommonFileFieldState();
}

class _CommonFileFieldState extends State<CommonFileField> {
  download() async {
    var extStorage =
    await getExternalStorageDirectories(type: StorageDirectory.downloads);
    if (extStorage.length > 0) {
      var name = '${extStorage[0].path}/${widget.fileInfo.name}';
      final file = File(name);
      file.writeAsBytesSync(widget.fileInfo.content);
      CommonConversion.showAlertDialog(
          context,
          'Info',
          'file $name with ${widget.fileInfo.content.length} bytes saved with success',
              () {});
    }
  }

  upload() async {
    File file = await FilePicker.getFile();
    if (file != null) {
      List<int> imageBytes = file.readAsBytesSync();
      var posB = file.path.lastIndexOf('/');
      var name = file.path.substring(posB + 1);
      var posE = name.lastIndexOf('.');
      var extension = name.substring(posE + 1).toLowerCase();
      var type = '';
      if (extension == 'jpg')
        type = 'image/jpeg';
      else if (extension == 'png')
        type = 'image/png';
      else if (extension == 'gif')
        type = 'image/gif';
      else if (extension == 'bmp')
        type = 'image/bmp';
      else if (extension == 'mp3')
        type = 'audio/mpeg';
      else if (extension == 'wav')
        type = 'audio/wav';
      else if (extension == 'txt')
        type = 'text/plain';
      else if (extension == 'pdf')
        type = 'application/pdf';
      else
        type = 'application/$extension';
      String base64Image = base64Encode(imageBytes);
      widget.onSelect(name,
          '{"file":{"name":"$name", "size":${imageBytes.length}, "type":"$type"}, "image":"data:$type;base64,$base64Image"}');
    }
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    if (widget.fileInfo != null) {
      var downloadBtn = Wrap(
        children: <Widget>[
          CommonIconAndTextButton(
            'Download ${widget.fileInfo.name}',
            FontAwesomeIcons.download,
            DefaultTheme.buttomDeleteFNColor,
            DefaultTheme.buttomDeleteBGColor,
            download,
          ),
        ],
      );
      if (widget.fileInfo.type.indexOf('image') >= 0) {
        return Column(
          children: <Widget>[
            SingleChildScrollView(
              child: new Image.memory(
                widget.fileInfo.content,
                height: MediaQuery.of(context).size.height - 180,
              ),
            ),
            downloadBtn,
          ],
        );
      } else {
        return downloadBtn;
      }
    } else {
      return Column(
        children: <Widget>[
          Wrap(
            children: <Widget>[
              CommonIconAndTextButton(
                BeeVoterController.translate(widget.title),
                FontAwesomeIcons.upload,
                DefaultTheme.buttomDeleteFNColor,
                DefaultTheme.buttomDeleteBGColor,
                upload,
              ),
            ],
          ),
        ],
      );
    }
  }
}

class CommonField extends StatefulWidget {
  final BuildContext context;
  final String title;
  // final String value;
  final bool isPassword;
  final FormFieldSetter<String> onChanged;
  final FocusNode currentFocus;
  final FocusNode nextFocus;
  final TextInputType keyboardType;
  final int lines;
  final TextEditingController controller;

  CommonField(
      this.context,
      this.title,
      // this.value,
      this.controller,
      this.isPassword,
      this.onChanged,
      this.currentFocus,
      this.nextFocus,
      this.keyboardType, {
        this.lines = 1,
      });

  @override
  _CommonFieldState createState() => _CommonFieldState();
}

class _CommonFieldState extends State<CommonField> {
  // static int lastPos = -1;

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return TextFormField(
      obscureText: widget.isPassword, // Use secure text for passwords.
      decoration: new InputDecoration(
        hintText: BeeVoterController.translate(widget.title),
        hintStyle: CommonList12TextStyle(),
        fillColor: DefaultTheme.backgroundColor,
      ),
//          validator: validator,
      onChanged: (String value) {
        widget.onChanged(value);
      },
      cursorColor: DefaultTheme.textColor,
      keyboardType: widget.keyboardType,
      maxLines: widget.lines,
      controller: widget.controller,
      style: CommonList12TextStyle(),
      focusNode: widget.currentFocus,
      textInputAction: widget.nextFocus != null
          ? TextInputAction.next
          : TextInputAction.done,
      onFieldSubmitted: (term) {
        widget.currentFocus.unfocus();
        if (widget.nextFocus != null) // && widget.value.isNotEmpty)
          FocusScope.of(context).requestFocus(widget.nextFocus);
      },
    );
  }
}

class CommonFieldWithIcon extends Padding {
  CommonFieldWithIcon(
      IconData icon,
      String hintText,
      ValueChanged<String> onChanged,
      TextEditingController controller,
      TextInputType keyboardType,
      {bool isPassword = false,
        ValueChanged<String> onNextField,
        FocusNode focusNode})
      : super(
      padding: const EdgeInsets.only(bottom: 5.0, top: 5.0),
      child: Container(
        width: 250,
        child: Material(
          elevation: 5.0,
          borderRadius: BorderRadius.all(Radius.circular(10.0)),
          color: Colors.yellowAccent,
          child: Row(
            mainAxisAlignment: MainAxisAlignment.spaceBetween,
            children: <Widget>[
              Container(
                padding: const EdgeInsets.all(4.0),
                width: 45,
                child: icon == null
                    ? Container(
                  width: 0,
                )
                    : Icon(
                  icon,
                  color: Colors.black,
                ),
              ),
              Flexible(
                child: Container(
                  decoration: BoxDecoration(
                    color: Colors.white,
                    borderRadius: BorderRadius.only(
                        topRight: Radius.circular(10.0),
                        bottomRight: Radius.circular(10.0)),
                  ),
//                      width: 200,
                  child: Padding(
                    padding: const EdgeInsets.all(4.0),
                    child: TextFormField(
                      focusNode: focusNode,
                      onFieldSubmitted: onNextField,
                      textInputAction: TextInputAction.next,
                      obscureText: isPassword,
                      onChanged: onChanged,
                      controller: controller,
                      keyboardType: keyboardType,
                      decoration: InputDecoration(
                        border: InputBorder.none,
                        hintText: hintText,
                        fillColor: Colors.white,
                        filled: true,
                      ),
                      style: TextStyle(
                          fontSize: 12, color: Colors.orange[800]),
                    ),
                  ),
                ),
              ),
            ],
          ),
        ),
      ));
}

abstract class CommonPopupMenuItem {
  doAction();
  PopupMenuEntry<CommonPopupMenuItem> build();
}

abstract class CommonPopupMenuItemText extends CommonPopupMenuItem {
  CommonPopupMenuItemText({this.title, this.icon, this.action});

  Widget title;
  IconData icon;
  VoidCallback action;

  doAction() {
    this.action();
  }

  PopupMenuEntry<CommonPopupMenuItem> build() {
    return PopupMenuItem<CommonPopupMenuItem>(
      value: this,
      child: title,
    );
  }
}

class CommonPopupMenuDivider extends CommonPopupMenuItem {
  doAction() {}
  PopupMenuEntry<CommonPopupMenuItem> build() {
    return PopupMenuDivider(height: 10);
  }
}

class CommonPopupMenu extends StatefulWidget {
  final List<CommonPopupMenuItem> choices;
  final String tooltip;
  final Color color;
  final Color fontColor;
  final Color buttonColor;

  CommonPopupMenu(this.choices,
      {this.tooltip, this.fontColor, this.color, this.buttonColor});

  @override
  _CommonPopupMenuState createState() => _CommonPopupMenuState(choices);
}

class _CommonPopupMenuState extends State<CommonPopupMenu> {
  _CommonPopupMenuState(List<CommonPopupMenuItem> choices);

  static Set<_CommonPopupMenuState> openedItens =
  new Set<_CommonPopupMenuState>();

  finish() {
    for (var i = 1; i < openedItens.length; i++) Navigator.pop(context);
    openedItens.clear();
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return Container(
      decoration: widget.buttonColor == null
          ? null
          : BoxDecoration(color: widget.buttonColor),
      child: PopupMenuButton<CommonPopupMenuItem>(
        elevation: 3.2,
        color:
        widget.color == null ? DefaultTheme.backgroundColor : widget.color,
        icon: CommonIcon20(
          Icons.more_vert,
          color: widget.fontColor == null
              ? DefaultTheme.textColor
              : widget.fontColor,
        ),
        padding: EdgeInsets.all(1),
        initialValue: widget.choices[0],
        onCanceled: () {
          finish();
        },
        tooltip: widget.tooltip,
        onSelected: (CommonPopupMenuItem choice) {
          finish();
          choice.doAction();
        },
        itemBuilder: (BuildContext context) {
          if (!openedItens.contains(this)) openedItens.add(this);
          return widget.choices.map((CommonPopupMenuItem choice) {
            return choice.build();
          }).toList();
        },
      ),
    );
  }
}

class CommonPopupMenuText extends CommonPopupMenuItemText {
  CommonPopupMenuText(String title, VoidCallback action)
      : super(
      title: Text(title, style: CommonList10TextStyle()), action: action);
}

class CommonIconAndText extends Row {
  CommonIconAndText(IconData icon, String title)
      : super(
    children: <Widget>[
      Padding(
        padding: EdgeInsets.only(right: 5),
        child: CommonIcon12(
          icon,
        ),
      ),
      Flexible(
        child: Text(title, style: CommonList12TextStyle()),
      ),
    ],
  );
}

class CommonScaffoldIconAndTitle extends Container {
  CommonScaffoldIconAndTitle(IconData icon, String title)
      : super(
    child: Row(
      mainAxisAlignment: MainAxisAlignment.start,
//            mainAxisAlignment: MainAxisAlignment.spaceBetween,
//            mainAxisAlignment: MainAxisAlignment.spaceEvenly,
//            crossAxisAlignment: CrossAxisAlignment.center,
//            mainAxisSize: MainAxisSize.max,
      children: <Widget>[
        Padding(
          padding: EdgeInsets.only(right: 5),
          child:
          CommonIcon12(icon, color: DefaultTheme.barTitleFontColor),
        ),

        Flexible(
          child:
//                Wrap(alignment: WrapAlignment.spaceEvenly, children: <Widget>[

          Row(
              mainAxisAlignment: MainAxisAlignment.spaceBetween,
              children: <Widget>[
                Text(
                  title,
                  style: TextStyle(
                    color: DefaultTheme.barTitleFontColor,
                    fontSize: DefaultTheme.fontSmall,
                  ),
                ),
                Image(
                  image: AssetImage('assets/bee_logo.png'),
                  height: 20,
                )
              ]),
        ),

//              Expanded(
//                child: ,
//              ),
      ],
    ),
  );
}

class CommonTextAndIcon extends Row {
  CommonTextAndIcon(IconData icon, String title)
      : super(
    children: <Widget>[
      Padding(
        padding: EdgeInsets.only(right: 5),
        child: Text(title, style: CommonList12TextStyle()),
      ),
      CommonIcon12(icon),
    ],
  );
}

class CommonPopupMenuIconAndText extends CommonPopupMenuItemText {
  CommonPopupMenuIconAndText(IconData icon, String title, VoidCallback action)
      : super(title: CommonIconAndText(icon, title), action: action);
}

class CommonPopupMenuSubmenu extends CommonPopupMenuItemText {
  CommonPopupMenuSubmenu(String title, List<CommonPopupMenuItem> subitens)
      : super(
    title: Row(
      children: <Widget>[
        Padding(
          padding: EdgeInsets.only(right: 5),
          child: CommonPopupMenu(subitens),
        ),
        Flexible(
          child: Text(title, style: CommonList12TextStyle()),
        ),
      ],
    ),
  );
}

class CommonPopupMenuIconAndTextChecked extends CommonPopupMenuItemText {
  CommonPopupMenuIconAndTextChecked(
      IconData icon, String title, VoidCallback action)
      : super(
      title: Row(
        children: <Widget>[
          CommonIcon12(icon),
          Padding(
            padding: EdgeInsets.only(left: 5, right: 5),
            child: Text(title, style: CommonList12TextStyle()),
          ),
          CommonIcon12(Icons.check)
        ],
      ),
      action: action);
}

class CommonCrud extends StatefulWidget {
  final VoidCallback onAdd;
  final void Function(int index) onEdit;
  final void Function(int index) onDelete;
  final int length;
  final String Function(int index) readItem;
  final Widget Function(int index) readItemWidget;

  CommonCrud(this.length, this.readItem,
      {this.onAdd, this.onEdit, this.onDelete, this.readItemWidget});

  @override
  _CommonCrudState createState() => _CommonCrudState();
}

class _CommonCrudState extends State<CommonCrud> {
  Widget _getListItemTile(BuildContext context, int index) {
    Widget item;
    Widget text;
    if (widget.readItem != null)
      text = Text(widget.readItem(index), style: CommonList10TextStyle());
    else if (widget.readItemWidget != null) text = widget.readItemWidget(index);

    if (widget.onEdit == null && widget.onDelete == null) {
      item = text;
    } else {
      var subItens = []; // new List<Widget>();
      if (widget.onEdit != null)
        subItens.add(CommonIconButton(
          FontAwesomeIcons.solidEdit,
              () => {widget.onEdit(index)},
          backgroundColor: DefaultTheme.buttomAddBGColor,
          fontColor: DefaultTheme.buttomAddFNColor,
        ));
      if (widget.onDelete != null)
        subItens.add(CommonIconButton(
            FontAwesomeIcons.solidTrashAlt,
                () => {
              CommonConversion.showDeleteConfirm(
                  context, () => {widget.onDelete(index)})
            },
            backgroundColor: DefaultTheme.buttomDeleteBGColor,
            fontColor: DefaultTheme.buttomDeleteFNColor));
      item = Row(
        mainAxisAlignment: MainAxisAlignment.spaceBetween,
        children: <Widget>[
          Flexible(child: text),
          Row(
            children: subItens,
          )
        ],
      );
    }

    return GestureDetector(onTap: () {}, child: CommomListItem(index, item));
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    var cildren = []; // new List<Widget>();
    if (widget.onAdd != null)
      cildren.add(CommonIconButton(
        FontAwesomeIcons.solidPlusSquare,
        widget.onAdd,
        backgroundColor: DefaultTheme.buttomAddBGColor,
        fontColor: DefaultTheme.buttomAddFNColor,
      ));
    for (var i = 0; i < widget.length; i++)
      cildren.add(_getListItemTile(context, i));

    return Padding(
      padding: EdgeInsets.all(0),
      child: SingleChildScrollView(
        child: Column(
          crossAxisAlignment: CrossAxisAlignment.stretch,
          mainAxisAlignment: MainAxisAlignment.end,
          children: cildren,
        ),
      ),
    );
  }
}

class CommonWaitable extends StatelessWidget {
  static VoidCallback onClearError;
  static bool isWaiting = false;
  final Widget child;
  CommonWaitable(this.child);

  static setError(BuildContext context, dynamic err, VoidCallback action) {
    CommonConversion.showAlertDialog(context, 'Error', err.toString(), () {});
    onClearError = action;
    isWaiting = false;
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    if (isWaiting)
      return Container(
          color: DefaultTheme.backgroundColor,
          child: Center(
            child: CircularProgressIndicator(
              valueColor:
              new AlwaysStoppedAnimation<Color>(DefaultTheme.folderBGColor),
            ),
          ));
    return child;
  }
}

class CommonMultiSelect extends FormField<dynamic> {
  final String titleText;
  final String hintText;
  final bool required;
  final String errorText;
  final dynamic value;
  final bool filterable;
  final List dataSource;
  final String textField;
  final String valueField;
  final Function change;
  final Function open;
  final Function close;
  final Widget leading;
  final Widget trailing;
  final int maxLength;
  CommonMultiSelect(
      {FormFieldSetter<dynamic> onSaved,
        void Function(List values) onSelect,
        FormFieldValidator<dynamic> validator,
        dynamic initialValue,
        // bool autovalidate = false,
        this.titleText = 'Title',
        this.hintText = 'Tap to select one or more...',
        this.required = false,
        this.errorText = 'Please select one or more option(s)',
        this.value,
        this.leading,
        this.filterable = true,
        this.dataSource,
        this.textField,
        this.valueField,
        this.change,
        this.open,
        this.close,
        this.trailing,
        this.maxLength})
      : super(
      onSaved: onSaved,
      validator: validator,
      initialValue: initialValue,
      // autovalidate: autovalidate,
      builder: (FormFieldState<dynamic> state) {
        List<Widget> _buildSelectedOptions(dynamic values, state) {
          List<Widget> selectedOptions = [];

          if (values != null) {
            values.forEach((item) {
              var existingItem = dataSource.singleWhere(
                      (itm) => itm[valueField] == item,
                  orElse: () => null);
              if (existingItem != null) {
                selectedOptions.add(Chip(
                  backgroundColor: Colors.orange,
                  label: Text(existingItem[textField],
                      style: TextStyle(fontSize: DefaultTheme.fontSmall),
                      overflow: TextOverflow.ellipsis),
                ));
              }
            });
          }

          return selectedOptions;
        }

        List values = state.value ?? [];

        return InkWell(
            onTap: () async {
              var results = await Navigator.push(
                  state.context,
                  MaterialPageRoute<dynamic>(
                    builder: (BuildContext context) =>
                        CommonSelectionModal(
                            onSelect: onSelect,
                            title: titleText,
                            filterable: filterable,
                            valueField: valueField,
                            textField: textField,
                            dataSource: dataSource,
                            values: values,
                            maxLength: maxLength ?? dataSource?.length),
                    fullscreenDialog: true,
                  ));

              if (results != null) {
                dynamic newValue;
                if (results.length > 0) {
                  newValue = results;
                }
                state.didChange(newValue);
                if (change != null) {
                  change(newValue);
                }
              }
            },
            child: InputDecorator(
              decoration: InputDecoration(
                filled: true,
                fillColor: DefaultTheme.backgroundColor,
                contentPadding:
                EdgeInsets.only(left: 10.0, top: 0.0, right: 10.0),
                errorBorder: OutlineInputBorder(
                    borderRadius: BorderRadius.all(Radius.circular(5.0)),
                    borderSide: BorderSide(color: Colors.red.shade400)),
                enabledBorder: OutlineInputBorder(
                    borderRadius: BorderRadius.all(Radius.circular(5.0)),
                    borderSide: BorderSide(color: Colors.grey.shade400)),
                errorText: state.hasError ? state.errorText : null,
                errorMaxLines: 50,
              ),
              isEmpty: (state.value == null ||
                  state.value == '' ||
                  (state.value != null && state.value.length == 0)),
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.start,
                children: <Widget>[
                  Padding(
                    padding: const EdgeInsets.only(top: 8.0),
                    child: Row(
                      crossAxisAlignment: CrossAxisAlignment.start,
                      children: <Widget>[
                        Expanded(
                          child: RichText(
                            text: TextSpan(
                                text: titleText,
                                style: TextStyle(
                                    fontSize: DefaultTheme.fontSmall,
                                    color: Theme.of(state.context)
                                        .primaryColor),
                                children: [
                                  TextSpan(
                                    text: required ? ' *' : '',
                                    style: TextStyle(
                                        color: Colors.red,
                                        fontSize: DefaultTheme.fontLarge),
                                  ),
                                  TextSpan(
                                    text: maxLength != null
                                        ? '(max $maxLength)'
                                        : '',
                                    style: TextStyle(
                                        color: Colors.red,
                                        fontSize:
                                        DefaultTheme.fontXSmall),
                                  )
                                ]),
                          ),
                        ),
                        Column(
                          crossAxisAlignment: CrossAxisAlignment.center,
                          mainAxisAlignment: MainAxisAlignment.center,
                          mainAxisSize: MainAxisSize.max,
                          children: <Widget>[
                            CommonIcon20(
                              Icons.arrow_downward,
                              color: Theme.of(state.context).primaryColor,
                            )
                          ],
                        )
                      ],
                    ),
                  ),
                  (state.value == null ||
                      state.value == '' ||
                      (state.value != null &&
                          state.value.length == 0))
                      ? new Container(
                    margin:
                    EdgeInsets.fromLTRB(0.0, 10.0, 0.0, 6.0),
                    child: Text(
                      hintText,
                      style: TextStyle(
                        color: Colors.grey.shade500,
                        fontSize: DefaultTheme.fontSmall,
                      ),
                    ),
                  )
                      : Wrap(
                    spacing: 8.0, // gap between adjacent chips
                    runSpacing: 1.0, // gap between lines
                    children:
                    _buildSelectedOptions(state.value, state),
                  )
                ],
              ),
            ));
      });
}

class CommonSelectionModal extends StatefulWidget {
  final List dataSource;
  final List values;
  final bool filterable;
  final String textField;
  final String valueField;
  final String title;
  final int maxLength;
  final void Function(List values) onSelect;

  CommonSelectionModal(
      {this.onSelect,
        this.filterable,
        this.dataSource,
        this.title = 'Please select one or more option(s)',
        this.values,
        this.textField,
        this.valueField,
        this.maxLength})
      : super();

  @override
  _CommonSelectionModalState createState() => _CommonSelectionModalState();
}

class _CommonSelectionModalState extends State<CommonSelectionModal> {
  Icon icon = new Icon(
    Icons.search,
    color: DefaultTheme.backgroundColor,
  );
  final globalKey = new GlobalKey<ScaffoldState>();
  final TextEditingController _controller = new TextEditingController();
  bool _isSearching;

  List _localDataSourceWithState = [];
  List _searchresult = [];

  _CommonSelectionModalState() {
    _controller.addListener(() {
      if (_controller.text.isEmpty) {
        setState(() {
          _isSearching = false;
        });
      } else {
        setState(() {
          _isSearching = true;
        });
      }
    });
  }

  @override
  void initState() {
    super.initState();
    widget.dataSource.forEach((item) {
      var newItem = {
        'value': item[widget.valueField],
        'text': item[widget.textField],
        'checked': widget.values.contains(item[widget.valueField])
      };
      _localDataSourceWithState.add(newItem);
    });

    _searchresult = List.from(_localDataSourceWithState);
    _isSearching = false;
  }

  Widget _buildAppBar(BuildContext context) {
    return AppBar(
      title: Text(
        widget.title,
        style: TextStyle(
          color: DefaultTheme.barTitleFontColor,
          fontSize: DefaultTheme.fontSmall,
        ),
      ),
    );
  }

  Widget _buildBody(BuildContext context) {
    return SafeArea(
      child: Column(
        children: <Widget>[
          widget.filterable ? _buildSearchText() : new SizedBox(),
          Expanded(
            child: _optionsList(),
          ),
          _currentlySelectedOptions(),
          Row(
            crossAxisAlignment: CrossAxisAlignment.end,
//          mainAxisAlignment: MainAxisAlignment.center,
            mainAxisAlignment: MainAxisAlignment.end,
            children: <Widget>[
              CommonIconButton(
                FontAwesomeIcons.shareSquare,
                widget.maxLength != null &&
                    _localDataSourceWithState
                        .where((item) => item['checked'])
                        .length >
                        widget.maxLength
                    ? null
                    : () {
                  var selectedValuesObjectList = _localDataSourceWithState
                      .where((item) => item['checked'])
                      .toList();
                  var selectedValues = [];
                  selectedValuesObjectList.forEach((item) {
                    selectedValues.add(item['value']);
                  });
                  Navigator.pop(context, selectedValues);

                  if (widget.onSelect != null)
                    widget.onSelect(selectedValues);
                },
                tooltip: 'Save',
              ),
            ],
          ),
        ],
      ),
    );
  }

  Widget _currentlySelectedOptions() {
    List<Widget> selectedOptions = [];

    var selectedValuesObjectList =
    _localDataSourceWithState.where((item) => item['checked']).toList();
    var selectedValues = [];
    selectedValuesObjectList.forEach((item) {
      selectedValues.add(item['value']);
    });
    selectedValues.forEach((item) {
      var existingItem = _localDataSourceWithState
          .singleWhere((itm) => itm['value'] == item, orElse: () => null);
      selectedOptions.add(Chip(
        label: new Container(
          constraints: new BoxConstraints(
              maxWidth: MediaQuery.of(context).size.width - 80.0),
          child: Text(existingItem['text'], overflow: TextOverflow.ellipsis),
        ),
        deleteButtonTooltipMessage: 'Tap to delete this item',
        deleteIcon: CommonIcon20(
          Icons.cancel,
          color: Colors.grey[700],
        ),
        deleteIconColor: Colors.grey,
        onDeleted: () {
          existingItem['checked'] = false;
          setState(() {});
        },
      ));
    });
    return selectedOptions.length > 0
        ? Container(
      padding: EdgeInsets.all(10.0),
      color: Colors.grey.shade400,
      child: Column(
        mainAxisAlignment: MainAxisAlignment.start,
        crossAxisAlignment: CrossAxisAlignment.stretch,
        children: <Widget>[
          new Text(
            'Currently selected ${selectedOptions.length} items (tap to remove)', // use languageService here
            style: TextStyle(
                color: Colors.black87, fontWeight: FontWeight.bold),
          ),
          ConstrainedBox(
              constraints: new BoxConstraints(
                maxHeight: MediaQuery.of(context).size.height / 8,
              ),
              child: Scrollbar(
                child: SingleChildScrollView(
                    child: Wrap(
                      spacing: 8.0, // gap between adjacent chips
                      runSpacing: 0.4, // gap between lines
                      alignment: WrapAlignment.start,
                      children: selectedOptions,
                    )),
              )),
        ],
      ),
    )
        : new Container();
  }

  ListView _optionsList() {
    List<Widget> options = [];
    _searchresult.forEach((item) {
      options.add(ListTile(
          title: Text(
            item['text'] ?? '',
            style: TextStyle(
              color: DefaultTheme.textColor,
              fontSize: DefaultTheme.fontSmall,
            ),
          ),
          leading: Transform.scale(
            scale: 1.0,
            child: CommonIcon20(
                item['checked']
                    ? Icons.check_box
                    : Icons.check_box_outline_blank,
                color: Theme.of(context).primaryColor),
          ),
          onTap: () {
            item['checked'] = !item['checked'];
            setState(() {});
          }));
      options.add(new Divider(height: 1.0));
    });
    return ListView(children: options);
  }

  Widget _buildSearchText() {
    return Container(
      color: Theme.of(context).primaryColor,
      padding: EdgeInsets.only(left: 8.0, right: 8.0, bottom: 10.0),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.stretch,
        children: <Widget>[
          Container(
            child: TextField(
              controller: _controller,
              keyboardAppearance: Brightness.light,
              onChanged: searchOperation,
              decoration: new InputDecoration(
                  contentPadding: EdgeInsets.all(12.0),
                  border: new OutlineInputBorder(
                    borderRadius: const BorderRadius.all(
                      const Radius.circular(6.0),
                    ),
                  ),
                  filled: true,
                  hintText: "Search...",
                  fillColor: DefaultTheme.backgroundColor,
                  suffix: SizedBox(
                      height: 25.0,
                      child: IconButton(
                        icon: Icon(Icons.clear),
                        onPressed: () {
                          _controller.clear();
                          searchOperation('');
                        },
                        padding: EdgeInsets.all(0.0),
                        tooltip: 'Clear',
                      ))),
            ),
          ),
        ],
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return new Scaffold(
      key: globalKey,
      appBar: _buildAppBar(context),
      body: _buildBody(context),
    );
  }

  void searchOperation(String searchText) {
    _searchresult.clear();
    if (_isSearching != null &&
        searchText != null &&
        searchText.toString().trim() != '') {
      for (int i = 0; i < _localDataSourceWithState.length; i++) {
        String data =
            '${_localDataSourceWithState[i]['value']} ${_localDataSourceWithState[i]['text']}';
        if (data.toLowerCase().contains(searchText.toLowerCase())) {
          _searchresult.add(_localDataSourceWithState[i]);
        }
      }
    } else {
      _searchresult = List.from(_localDataSourceWithState);
    }
  }
}

class CommonFormContent extends Expanded {
  CommonFormContent(List<Widget> fields)
      : super(
      child: ListView(
        scrollDirection: Axis.vertical,
        children: fields,
      ));
}

class CommonForm extends StatelessWidget {
  final List<Widget> fields;
  final List<Widget> extraActions;
  final VoidCallback onSave;
  final bool isOk;
  final VoidCallback onDelete;

  CommonForm(this.fields, this.isOk, this.onSave,
      {this.onDelete, this.extraActions});

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    var actions = []; // new List<Widget>();
    if (onSave != null)
      actions.add(CommonIconButton(
        FontAwesomeIcons.shareSquare,
        onSave,
        tooltip: BeeVoterController.translate('Save'),
        fontColor: DefaultTheme.buttomSaveFNColor,
        backgroundColor: DefaultTheme.buttomSaveBGColor,
        isEnabled: isOk,
      ));
    if (onDelete != null)
      actions.add(CommonIconButton(
        FontAwesomeIcons.solidTrashAlt,
            () {
          CommonConversion.showDeleteConfirm(context, () async {
            onDelete();
          });
        },
        tooltip: BeeVoterController.translate('Delete'),
        fontColor: DefaultTheme.buttomDeleteFNColor,
        backgroundColor: DefaultTheme.buttomDeleteBGColor,
        // isEnabled: !widget.isNew,
      ));
    if (extraActions != null) actions.addAll(extraActions);
    return Container(
      color: DefaultTheme.backgroundColor,
      child: CommonWaitable(
        Padding(
          padding: EdgeInsets.all(5),
          child: Container(
            width: double.infinity,
            child: Column(
              children: <Widget>[
                CommonFormContent(fields),
                Row(
                  crossAxisAlignment: CrossAxisAlignment.end,
                  mainAxisAlignment: MainAxisAlignment.end,
                  children: actions,
                ),
              ],
            ),
          ),
        ),
      ),
    );
  }
}

class CommonExpandableController {
  ExpandableController currentOpened;
  int currentOpenedId;
}

class CommonExpandablePanel extends StatelessWidget {
  final int id;
  final CommonExpandableController controller;
  final IconData icon;
  final String title;

  // final Widget header;
  // final Widget collapsed;
  final Widget child;
  // final bool tapHeaderToExpand;
  // final bool hasIcon;
  final bool visible;

  // this.controller,
  // this.tapBodyToCollapse = false,
  // this.builder,
  // ExpandableThemeData theme,

  CommonExpandablePanel(
      this.id,
      this.controller,
      this.icon,
      this.title, {
        // this.header,
        // this.collapsed,
        this.child,
        // this.tapHeaderToExpand = true,
        // this.hasIcon = true,
        this.visible = true,
      });

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    ExpandableController ctrl = ExpandableController();
    ctrl.addListener(() {
      if (controller.currentOpened != null && controller.currentOpened != ctrl)
        controller.currentOpened.toggle();
      controller.currentOpened = ctrl;
      controller.currentOpenedId = id;
    });

    if (controller.currentOpenedId == id) ctrl.toggle();

    var exp = ExpandableNotifier(
      initialExpanded: false,
      child: ScrollOnExpand(
        theme: ExpandableThemeData(
          animationDuration: Duration(milliseconds: 500),
        ),
        child: ExpandablePanel(
            header:
            CommonIconAndText(icon, BeeVoterController.translate(title)),
            controller: ctrl,
            theme: ExpandableThemeData(
              iconColor: DefaultTheme.textColor,
            ),
            collapsed: Container(height: 0.0), //collapsed,
            expanded: child,
            // tapHeaderToExpand: tapHeaderToExpand,
            // hasIcon: hasIcon
        ),
      ),
    );

    return visible
        ? exp
        : Container(
      width: 0,
    );
  }
}

class CommomListItem extends Container {
  CommomListItem(int index, Widget title, {Color color, Color borderColor})
      : super(
    child: ListTile(
      title: title,
    ),
    decoration: BoxDecoration(
      color: color==null ? (index % 2 == 0
          ? DefaultTheme.destakListColor
          : DefaultTheme.backgroundColor) : color,
      border: Border(
        bottom: BorderSide(
          color: borderColor == null ? (index % 2 == 0
              ? DefaultTheme.destakListColor
              : DefaultTheme.backgroundColor) : borderColor.withOpacity(0.2),
          width: 1.0,
        ),
      ),
    ),
  );
}

class CommomListItemText extends CommomListItem {
  CommomListItemText(int index, String title)
      : super(
    index,
    Text(title, style: CommonList12TextStyle()),
  );
}

class CommonAppBar extends AppBar {
  CommonAppBar(IconData icon, String title,
      {GestureTapCallback onTap, PreferredSizeWidget bottom})
      : super(
    iconTheme: IconThemeData(
      color: DefaultTheme.barTitleFontColor, //change your color here
    ),
    bottom: bottom,
    title: onTap != null
        ? GestureDetector(
        onTap: onTap,
        child: CommonScaffoldIconAndTitle(
            icon, BeeVoterController.translate(title)))
        : CommonScaffoldIconAndTitle(
        icon, BeeVoterController.translate(title)),
  );
}

class CommonRefreshable extends StatelessWidget {
  final RefreshController refreshController =
  RefreshController(initialRefresh: false);
  final Widget child;
  final Future Function() onRefresh;

  CommonRefreshable(this.onRefresh, this.child);

  refresh() async {
    await onRefresh();
    refreshController.refreshCompleted();
  }

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    return SmartRefresher(
      enablePullDown: true,
      enablePullUp: false,
      header: WaterDropHeader(
        waterDropColor: DefaultTheme.menuColor,
      ),
      controller: refreshController,
      onRefresh: refresh,
      onLoading: () => {print('loading')},
      child: child,
    );
  }
}

class CommonScaffold extends StatefulWidget {
  final IconData icon;
  final String title;
  final Widget body;
  final GestureTapCallback onTap;
  final Widget drawer;
  final Widget floatingActionButton;

  CommonScaffold(
      this.icon,
      this.title,
      this.body, {
        this.onTap,
        this.drawer,
        this.floatingActionButton,
      });

  @override
  _CommonScaffoldState createState() => _CommonScaffoldState();
}

class _CommonScaffoldState extends State<CommonScaffold> {
  GlobalKey<ScaffoldState> scaffoldKey = new GlobalKey();

  @override
  Widget build(BuildContext context) {
    BeeVoterController.lastContext = context;
    // if (widget.drawer == null && widget.floatingActionButton == null) {
    //   return Scaffold(
    //     body: CustomScrollView(
    //       slivers: <Widget>[
    //         SliverAppBar(
    //           leading: IconButton(
    //               icon: CommonIcon20(Icons.arrow_back,
    //                   color: DefaultTheme.barTitleFontColor),
    //               onPressed: () {
    //                 Navigator.of(context).pop(); // dismiss dialog
    //               }),
    //           expandedHeight: 10.0,
    //           floating: true,
    //           pinned: false,
    //           snap: false,
    //           elevation: 50,
    //           backgroundColor: DefaultTheme.barTitleBGColor,
    //           flexibleSpace: FlexibleSpaceBar(
    //             title: CommonScaffoldIconAndTitle(
    //                 widget.icon, BeeVoterController.translate(widget.title)),
    //           ),
    //         ),
    //         // widget.body,
    //         new SliverList(
    //             delegate: new SliverChildListDelegate(_buildList(50))),
    //       ],
    //     ),
    //   );
    // }

    return Scaffold(
      key: scaffoldKey,
      backgroundColor: CommonBeeVoterScaffold.bColor,
      appBar: CommonAppBar(
        widget.icon,
        widget.title,
        onTap: widget.onTap,
      ),
      body: widget.body,
      drawer: widget.drawer,
      drawerScrimColor:
      widget.drawer == null ? null : DefaultTheme.barTitleFontColor,
      floatingActionButton: widget.floatingActionButton,
    );
  }

// List _buildList(int count) {
//   List<Widget> listItems = List();

//   for (int i = 0; i < count; i++) {
//     listItems.add(new Padding(
//         padding: new EdgeInsets.all(20.0),
//         child: new Text('Item ${i.toString()}',
//             style: new TextStyle(fontSize: 25.0))));
//   }

//   return listItems;
// }
}
